/*
 * Common LCD device support for the JZ4740 and related SoCs.
 *
 * Copyright (C) 2018, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/lcd-jz4740.h>
#include <l4/devices/lcd-jz4780.h>
#include <l4/devices/panel-loader.h>
#include <l4/devices/memory.h>
#include "lcd-jz4740-device.h"

#include <l4/re/dataspace>
#include <l4/re/env>
#include <l4/re/env.h>
#include <l4/re/mem_alloc>
#include <l4/re/rm>
#include <l4/re/util/cap_alloc>
#include <l4/sys/types.h>

#include <stdint.h>

#include "activation_client.h"
#include "cpm_client.h"

// Virtual addresses for the LCD register block.

static l4_addr_t lcd_virt_base = 0, lcd_virt_base_end = 0;

// LCD, CPM and display device abstractions.

static Lcd_jz4740_chip *lcd_chip = 0;

static Activation *display_device;
static CPM *cpm_device;



// CPM operations.

static void set_timing()
{
  uint32_t pclk = lcd_chip->get_pixel_clock();

  cpm_device->stop_clock(Clock_lcd);

  // Original comment: LCDClock > 2.5*Pixclock
  // However, the documentation indicates that a TFT panel needs a device clock
  // 1.5 times that of the pixel clock, and a STN panel needs a device clock 3
  // times that of the pixel clock.

  cpm_device->set_frequency(Clock_frequency_lcd, pclk * 3);
  cpm_device->set_frequency(Clock_frequency_lcd_pixel, pclk);
  cpm_device->update_output_frequency();
  cpm_device->start_clock(Clock_lcd);

  l4_sleep(1); // 1ms == 1000us
}



// Disable the display.

void
Lcd_jz4740_device::disable()
{
  Lcd_jz4740_chip *chip = static_cast<Lcd_jz4740_chip *>(_chip);

  disable_display();
  chip->disable();
}

// Configure the peripheral and enable the display.

void
Lcd_jz4740_device::enable()
{
  Lcd_jz4740_chip *chip = static_cast<Lcd_jz4740_chip *>(_chip);

  // Initialise the clocks for the LCD controller. Without this, the controller
  // cannot be operated.

  set_timing();

  // Configure the controller.

  chip->disable();
  chip->config((struct Jz4740_lcd_descriptor *) desc_vaddr,
               (struct Jz4740_lcd_descriptor *) desc_paddr,
               fb_paddr);

  // Activate the display channel.

  enable_display();

  // Activate the controller output.

  chip->enable();
}

// Set up memory addresses for the peripheral, initialising the framebuffer and
// descriptor members.

int
Lcd_jz4740_device::_setup_memory()
{
  // Framebuffer and descriptor sizes.

  unsigned long fb_size, desc_size;

  // Size of physically contiguous regions to be allocated.

  l4_size_t fb_size_out, desc_size_out;

  // Memory allocation capability.

  L4::Cap<L4Re::Dataspace> descmem;

  // Test for existing setup.

  if (fb_vaddr)
    return 0;

  // Obtain capabilities for allocating memory.

  _fbmem = L4Re::Util::cap_alloc.alloc<L4Re::Dataspace>();
  if (!_fbmem.is_valid()) return 1;

  descmem = L4Re::Util::cap_alloc.alloc<L4Re::Dataspace>();
  if (!descmem.is_valid()) return 1;

  // Obtain the memory requirements.

  Lcd_jz4740_chip *chip = static_cast<Lcd_jz4740_chip *>(_chip);

  fb_size = chip->get_screen_size();
  desc_size = chip->get_descriptors_size();

  // Allocate memory for the framebuffer at 2**6 == 64 byte == 16 word alignment,
  // also for the descriptors.
  // NOTE: A 64 word burst mode is available on the JZ4780 that would
  // NOTE: necessitate 64 word alignment for the framebuffer.

  const l4_size_t alloc_flags = L4Re::Mem_alloc::Continuous | L4Re::Mem_alloc::Pinned;

  if (L4Re::Env::env()->mem_alloc()->alloc(fb_size, _fbmem, alloc_flags, 8) ||
      L4Re::Env::env()->mem_alloc()->alloc(desc_size, descmem, alloc_flags, 6))
    return 1;

  // Map the allocated memory, obtaining virtual addresses.

  const l4_size_t attach_flags = L4Re::Rm::Search_addr | L4Re::Rm::Eager_map;

  fb_vaddr = 0;
  desc_vaddr = 0;

  if (L4Re::Env::env()->rm()->attach(&fb_vaddr, fb_size, attach_flags, _fbmem, 0) ||
      L4Re::Env::env()->rm()->attach(&desc_vaddr, desc_size, attach_flags, descmem, 0))
    return 1;

  // Obtain physical addresses for the framebuffer and descriptors.

  fb_paddr = 0;
  desc_paddr = 0;

  fb_size_out = fb_size;
  desc_size_out = desc_size;

  if (_fbmem->phys(0, fb_paddr, fb_size_out) ||
      descmem->phys(0, desc_paddr, desc_size_out))
    return 1;

  // Test the mapped region sizes.

  if ((fb_size_out != fb_size) || (desc_size_out != desc_size))
    return 1;

  return 0;
}

// Populate a view information structure.

void
Lcd_jz4740_device::get_view_info(l4re_video_view_info_t *view_info)
{
  // Populate the L4Re framebuffer description with details from the
  // somewhat similar panel description.

  Lcd_jz4740_chip *chip = static_cast<Lcd_jz4740_chip *>(_chip);
  struct Jz4740_lcd_panel *panel = chip->get_panel();

  view_info->width                      = panel->width;
  view_info->height                     = panel->height;
  view_info->pixel_info.bytes_per_pixel = (panel->bpp + 7) / 8;
  view_info->bytes_per_line             = chip->get_line_size();

  switch (panel->bpp)
  {
    // NOTE: 24bpp here is actually RGBA or RGB padded.

    case 24:
    view_info->pixel_info.bytes_per_pixel = 4;
    view_info->pixel_info.r.shift = 16; view_info->pixel_info.r.size = 8;
    view_info->pixel_info.g.shift = 8;  view_info->pixel_info.g.size = 8;
    view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 8;
    break;

    case 32:
    view_info->pixel_info.r.shift = 16; view_info->pixel_info.r.size = 8;
    view_info->pixel_info.g.shift = 8;  view_info->pixel_info.g.size = 8;
    view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 8;
    break;

    case 8:
    view_info->pixel_info.r.shift = 5;  view_info->pixel_info.r.size = 3;
    view_info->pixel_info.g.shift = 2;  view_info->pixel_info.g.size = 3;
    view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 2;
    break;

    case 4:
    view_info->pixel_info.r.shift = 3;  view_info->pixel_info.r.size = 1;
    view_info->pixel_info.g.shift = 1;  view_info->pixel_info.g.size = 2;
    view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 1;
    break;

    case 16:
    default:
    view_info->pixel_info.r.shift = 11; view_info->pixel_info.r.size = 5;
    view_info->pixel_info.g.shift = 5;  view_info->pixel_info.g.size = 6;
    view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 5;
    break;
  }

  view_info->pixel_info.a.shift = 0;
  view_info->pixel_info.a.size  = 0;
}

// Access to peripheral memory.

static int setup_memory()
{
  bool is_jz4780 = false;

  if (get_memory("jz4740-lcd", &lcd_virt_base, &lcd_virt_base_end))
  {
    if (get_memory("jz4780-lcd", &lcd_virt_base, &lcd_virt_base_end))
      return 1;
    else
      is_jz4780 = true;
  }

  // Obtain access to the CPM and display devices.

  l4_cap_idx_t cpm = l4re_env_get_cap("cpm");
  if (!l4_is_valid_cap(cpm)) return 1;

  l4_cap_idx_t display = l4re_env_get_cap("display");
  if (!l4_is_valid_cap(display)) return 1;

  static client_Activation display_obj(display);
  display_device = &display_obj;

  static client_CPM cpm_obj(cpm);
  cpm_device = &cpm_obj;

  // Load the panel data from the configured library.

  struct Jz4740_lcd_panel *panel = (struct Jz4740_lcd_panel *) load_panel();

  // Initialise the LCD abstraction.

  if (is_jz4780)
    lcd_chip = new Lcd_jz4780_chip(lcd_virt_base, panel);
  else
    lcd_chip = new Lcd_jz4740_chip(lcd_virt_base, panel);

  return 0;
}

// Device initialisation.

Lcd_jz4740_device *Lcd_jz4740_device::device = 0;

Lcd_device *Lcd_device::get_device()
{
  if (Lcd_jz4740_device::device) return Lcd_jz4740_device::device;

  if (setup_memory()) return 0;

  // Initialise the common device.

  Lcd_jz4740_device::device = new Lcd_jz4740_device(lcd_chip, display_device);

  return Lcd_jz4740_device::device;
}
