/*
 * Clock and power management. This exposes the combined functionality
 * provided by the jz4730. The power management functionality could be exposed
 * using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-jz4730.h"



enum Regs : unsigned
{
  Clock_control       = 0x000,  // CFCR (CPCCR in JZ4740)
  Pll_control         = 0x010,  // PLCR1 (CPPCR in JZ4740)
  Clock_gate          = 0x020,  // MSCR (CLKGR in JZ4740)
  Sleep_control       = 0x024,  // SCR
  Lcd_divider         = 0x060,  // CFCR2
};

enum Clock_bits : unsigned
{
  Clock_enable          = 20, // UPE
  Clock_memory_divider  = 16, // MFR
  Clock_lcd_divider     = 12, // LFR
  Clock_pclock_divider  = 8,  // PFR (slow APB peripherals)
  Clock_hclock_divider  = 4,  // SFR (fast AHB peripherals)
  Clock_cpu_divider     = 0,  // IFR
};

enum Pll_bits : unsigned
{
  Pll_multiplier        = 23, // PLL1FD
  Pll_input_division    = 18, // PLL1RD
  Pll_output_division   = 16, // PLL1OD
  Pll_stable            = 10, // PLL1S
  Pll_bypassed          = 9,  // PLL1BP
  Pll_enabled           = 8,  // PLL1EN
};

enum Clock_gate_bits : unsigned
{
  Clock_gate_uprt       = 25,
  Clock_gate_udc        = 24,
  Clock_gate_cim        = 23,
  Clock_gate_kbc        = 22,
  Clock_gate_emac       = 21,
  Clock_gate_uart3      = 20,
  Clock_gate_aic_bitclk = 18,
  Clock_gate_scc        = 14,
  Clock_gate_msc        = 13,
  Clock_gate_ssi        = 12,
  Clock_gate_pwm1       = 11,
  Clock_gate_pmw0       = 10,
  Clock_gate_aic_pclk   = 9,
  Clock_gate_i2c        = 8,
  Clock_gate_lcd        = 7,
  Clock_gate_uhc        = 6,
  Clock_gate_dmac       = 5,
  Clock_gate_timer      = 3,
  Clock_gate_uart2      = 2,
  Clock_gate_uart1      = 1,
  Clock_gate_uart0      = 0,
};

enum Lcd_divider_bits : unsigned
{
  Lcd_divider_value     = 0,  // PIXFR (in CFCR2)
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_jz4730_chip::Cpm_jz4730_chip(l4_addr_t addr, uint32_t exclk_freq)
: _exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // add_cid("cpm");
  // add_cid("cpm-jz4730");
  // register_property("exclk_freq", &_exclk_freq);
}

// Clock/timer control.

uint32_t
Cpm_jz4730_chip::get_clock_gate_value(enum Clock_identifiers clock)
{ 
  switch (clock)
  {
    case Clock_uprt: return (1 << Clock_gate_uprt);
    case Clock_udc: return (1 << Clock_gate_udc);
    case Clock_cim: return (1 << Clock_gate_cim);
    case Clock_kbc: return (1 << Clock_gate_kbc);
    case Clock_emac: return (1 << Clock_gate_emac);
    case Clock_uart3: return (1 << Clock_gate_uart3);
    case Clock_aic_bitclk: return (1 << Clock_gate_aic_bitclk);
    case Clock_scc: return (1 << Clock_gate_scc);
    case Clock_msc: return (1 << Clock_gate_msc);
    case Clock_ssi: return (1 << Clock_gate_ssi);
    case Clock_pwm1: return (1 << Clock_gate_pwm1);
    case Clock_pmw0: return (1 << Clock_gate_pmw0);
    case Clock_aic_pclk: return (1 << Clock_gate_aic_pclk);
    case Clock_i2c: return (1 << Clock_gate_i2c);
    case Clock_lcd: return (1 << Clock_gate_lcd);
    case Clock_uhc: return (1 << Clock_gate_uhc);
    case Clock_dma: return (1 << Clock_gate_dmac);
    case Clock_timer: return (1 << Clock_gate_timer);
    case Clock_uart2: return (1 << Clock_gate_uart2);
    case Clock_uart1: return (1 << Clock_gate_uart1);
    case Clock_uart0: return (1 << Clock_gate_uart0);
    default: return 0;
  }
}

int
Cpm_jz4730_chip::have_clock(enum Clock_identifiers clock)
{
  return !(_regs[Clock_gate] & get_clock_gate_value(clock));
}

void
Cpm_jz4730_chip::start_clock(enum Clock_identifiers clock)
{
  _regs[Clock_gate] = _regs[Clock_gate] & ~get_clock_gate_value(clock);
}

void
Cpm_jz4730_chip::stop_clock(enum Clock_identifiers clock)
{
  _regs[Clock_gate] = _regs[Clock_gate] | get_clock_gate_value(clock);
}

// PLL control.

// Return whether the PLL is stable.

int
Cpm_jz4730_chip::have_pll()
{
  return _regs[Pll_control] & (1 << Pll_stable);
}

int
Cpm_jz4730_chip::pll_enabled()
{
  return _regs[Pll_control] & (1 << Pll_enabled);
}

int
Cpm_jz4730_chip::pll_bypassed()
{
  return _regs[Pll_control] & (1 << Pll_bypassed);
}

// Feedback (9-bit) multiplier.

uint16_t
Cpm_jz4730_chip::get_multiplier()
{
  return ((_regs[Pll_control] & (0x1ff << Pll_multiplier)) >> Pll_multiplier) + 2;
}

// Input (5-bit) divider.

uint8_t
Cpm_jz4730_chip::get_input_division()
{
  return ((_regs[Pll_control] & (0x1f << Pll_input_division)) >> Pll_input_division) + 2;
}

// Output divider.

static uint8_t od[4] = {1, 2, 2, 4};

uint8_t
Cpm_jz4730_chip::get_output_division()
{
  return od[(_regs[Pll_control] & (0x03 << Pll_output_division)) >> Pll_output_division];
}

// General clock divider.

static uint8_t cd[] = {1, 2, 3, 4, 6, 8, 12, 16, 24, 32};

uint8_t
Cpm_jz4730_chip::_get_divider(uint32_t reg, uint32_t mask, uint8_t shift)
{
  uint8_t d = (_regs[reg] & mask) >> shift;
  return (d < 10) ? cd[d] : 1;
}

// CPU clock (CCLK) divider.

uint8_t
Cpm_jz4730_chip::get_cpu_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_cpu_divider, Clock_cpu_divider);
}

// Fast peripheral clock (HCLK) divider.

uint8_t
Cpm_jz4730_chip::get_hclock_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_hclock_divider, Clock_hclock_divider);
}

// Slow peripheral clock (PCLK) divider.

uint8_t
Cpm_jz4730_chip::get_pclock_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_pclock_divider, Clock_pclock_divider);
}

// Memory clock (MCLK) divider.

uint8_t
Cpm_jz4730_chip::get_memory_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_memory_divider, Clock_memory_divider);
}

// Clock source divider for MSC, I2S, LCD and USB.

uint8_t
Cpm_jz4730_chip::get_source_divider()
{
  return 1;
}

// LCD device clock divider.

void
Cpm_jz4730_chip::set_lcd_device_divider(uint8_t division)
{
  if (division == 0)
    division = 1;

  // NOTE: The vendor code (clock.c) bounds the divider at 16, but maybe this is
  // NOTE: confused with the width of the bitfield which actually contains an
  // NOTE: index for the clock divider value array (cd).

  else if (division > 16)
    division = 16;

  // Obtain the divider or closest higher divider.

  int i;

  for (i = 0; i < 10; i++)
    if (cd[i] >= division) break;

  _regs[Clock_control] = (_regs[Clock_control] & ~(0xf << Clock_lcd_divider)) |
                         (cd[i] << Clock_lcd_divider);
}

// LCD pixel clock divider.

uint16_t
Cpm_jz4730_chip::get_lcd_pixel_divider()
{
  return (_regs[Lcd_divider] >> Lcd_divider_value) + 1;
}

void
Cpm_jz4730_chip::set_lcd_pixel_divider(uint16_t division)
{
  if (division == 0)
    division = 1;
  else if (division > 512)
    division = 512;

  _regs[Lcd_divider] = (_regs[Lcd_divider] & ~(0x1ff << Lcd_divider_value)) |
                       ((division - 1) << Lcd_divider_value);
}



uint32_t
Cpm_jz4730_chip::get_frequency(enum Clock_frequency_identifiers clock)
{
  if (clock == Clock_frequency_lcd_pixel)
    return get_output_frequency() / get_lcd_pixel_divider();

  // NOTE: Consider a better error result.
  return 0;
}

void
Cpm_jz4730_chip::set_frequency(enum Clock_frequency_identifiers clock, uint32_t frequency)
{
  uint32_t out = get_output_frequency();

  switch (clock)
  {
    // Limit the device frequency to 150MHz.

    case Clock_frequency_lcd:
      if (frequency > 150000000)
        frequency = 150000000;
      set_lcd_device_divider(out / frequency);
      break;

    case Clock_frequency_lcd_pixel:
      set_lcd_pixel_divider(out / frequency);
      break;

    default:
      break;
  }
}



uint32_t
Cpm_jz4730_chip::get_pll_frequency()
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled() && !pll_bypassed())
    return (_exclk_freq * get_multiplier()) /
           (get_input_division() * get_output_division());
  else
    return _exclk_freq;
}

// Clock frequency for MSC, I2S, LCD and USB.

uint32_t
Cpm_jz4730_chip::get_output_frequency()
{
  return get_pll_frequency() / get_source_divider();
}

void
Cpm_jz4730_chip::update_output_frequency()
{
  _regs[Clock_control] = _regs[Clock_control] | (1 << Clock_enable);
}

// Clock frequency for the CPU.

uint32_t Cpm_jz4730_chip::get_cpu_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_cpu_divider();
  else
    return _exclk_freq;
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_jz4730_chip::get_hclock_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_hclock_divider();
  else
    return _exclk_freq;
}

// Clock frequency for slow peripherals.

uint32_t
Cpm_jz4730_chip::get_pclock_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_pclock_divider();
  else
    return _exclk_freq;
}

// Clock frequency for the memory.

uint32_t
Cpm_jz4730_chip::get_memory_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_memory_divider();
  else
    return _exclk_freq;
}



// C language interface functions.

void
*jz4730_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 3.6864 MHz EXCLK frequency. */

  return (void *) new Cpm_jz4730_chip(cpm_base, 3686400);
}

int
jz4730_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->have_clock(clock);
}

void
jz4730_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->start_clock(clock);
}

void
jz4730_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->stop_clock(clock);
}

uint32_t
jz4730_cpm_get_cpu_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_cpu_frequency();
}

uint32_t
jz4730_cpm_get_hclock_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_hclock_frequency();
}

uint32_t
jz4730_cpm_get_output_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_output_frequency();
}

uint32_t
jz4730_cpm_get_pclock_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_pclock_frequency();
}

uint32_t
jz4730_cpm_get_memory_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_memory_frequency();
}

uint16_t
jz4730_cpm_get_lcd_pixel_divider(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_lcd_pixel_divider();
}

uint32_t
jz4730_cpm_get_frequency(void *cpm, enum Clock_frequency_identifiers clock)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_frequency(clock);
}

void
jz4730_cpm_set_frequency(void *cpm, enum Clock_frequency_identifiers clock, uint32_t frequency)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->set_frequency(clock, frequency);
}

void
jz4730_cpm_update_output_frequency(void *cpm)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->update_output_frequency();
}
