/*
 * I2C support for the JZ4730.
 *
 * Copyright (C) 2017, 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/cpm-jz4730.h>
#include <l4/devices/hw_mmio_register_block.h>



// State machine states.

enum I2c_jz4730_state
{
  I2c_jz4730_end = 0,
  I2c_jz4730_pre_start,
  I2c_jz4730_start_read,
  I2c_jz4730_perform_read,
  I2c_jz4730_perform_write,
  I2c_jz4730_stop_write
};



// I2C channel.

class I2c_jz4730_channel
{
private:
  Hw::Register_block<32> _regs;
  Cpm_jz4730_chip *_cpm;
  uint32_t _frequency;
  l4_cap_idx_t _irq=L4_INVALID_CAP;

public:
  I2c_jz4730_channel(l4_addr_t start, Cpm_jz4730_chip *cpm,
                     uint32_t frequency, l4_cap_idx_t irq);

  void disable();
  void enable();

  unsigned int read(uint8_t address, uint8_t buf[], unsigned int length);
  unsigned int write(uint8_t address, uint8_t buf[], unsigned int length);

protected:
  void set_frequency();

  // State machine activities.

  unsigned int _nread, _nwritten, _length, _limit;
  uint8_t _address;
  uint8_t *_buf;
  bool _read;

  void communicate();

  enum I2c_jz4730_state pre_start();
  enum I2c_jz4730_state start_read();
  enum I2c_jz4730_state perform_read();
  enum I2c_jz4730_state perform_write();
  enum I2c_jz4730_state stop_write();

  // Transaction control.

  bool set_address(uint8_t address, bool read);

  void clear_next();
  void send_next();
  void signal_last();

  void start();
  void stop();

  bool wait_for_irq(unsigned int timeout);

  // Specific status conditions.

  bool busy();
  bool data_valid();
  bool nack();
  bool transferred();
  bool transferring();
};

// I2C device control.

class I2c_jz4730_chip
{
private:
  l4_addr_t _start, _end;
  Cpm_jz4730_chip *_cpm;
  uint32_t _frequency;

public:
  I2c_jz4730_chip(l4_addr_t start, l4_addr_t end, Cpm_jz4730_chip *cpm,
                  uint32_t frequency);

  I2c_jz4730_channel *get_channel(uint8_t channel, l4_cap_idx_t irq);
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4730_i2c_init(l4_addr_t start, l4_addr_t end, void *cpm,
                      uint32_t frequency);

void *jz4730_i2c_get_channel(void *i2c, uint8_t channel, l4_cap_idx_t irq);

void jz4730_i2c_disable(void *i2c_channel);

void jz4730_i2c_enable(void *i2c_channel);

unsigned int jz4730_i2c_read(void *i2c_channel, uint8_t address, uint8_t buf[], unsigned int length);

unsigned int jz4730_i2c_write(void *i2c_channel, uint8_t address, uint8_t buf[], unsigned int length);

EXTERN_C_END
